import argparse
from argparse import ArgumentParser
import asyncio
from dataclasses import dataclass, fields
import os
from pathlib import Path
from typing import AsyncGenerator, Dict, List, Literal, Optional, Tuple, Union

from transformers import BatchEncoding, PreTrainedTokenizerBase
from typing_extensions import TypedDict

from furiosa.native_runtime.llm import NativeLLMEngine, NativeRequestOutput  # type: ignore
from furiosa_llm.api import (
    CACHE_DIR,
    LLM,
    LLMBackend,
    RequestOutput,
    SamplingParams,
    SchedulerConfig,
    TokenizerModeType,
)
from furiosa_llm.outputs import NativeOutputConverter, RequestOutputKind
from furiosa_llm.server.utils import AnyTokenizer


# adopted from https://github.com/vllm-project/vllm/blob/main/vllm/inputs/data.py
class TextPrompt(TypedDict):
    """Schema for a text prompt."""

    prompt: str
    """The input text to be tokenized before passing to the model."""


class TokensPrompt(TypedDict):
    """Schema for a tokenized prompt."""

    prompt_token_ids: List[int]
    """A list of token IDs to pass to the model."""


SingletonPrompt = Union[str, TextPrompt, TokensPrompt]

# TODO: support ExplicitEncoderDecoderPrompt later
PromptType = Union[SingletonPrompt]


@dataclass
class EngineArgs:
    # Currently only artifact path is supported
    model: str
    num_speculative_tokens: Optional[int] = None
    use_speculative_decoding_if_possible: bool = True
    data_parallel_size: Optional[int] = None
    tokenizer: Optional[str] = None
    tokenizer_mode: TokenizerModeType = "auto"
    seed: Optional[int] = None
    devices: Optional[str] = None
    cache_dir: os.PathLike = CACHE_DIR
    backend: Optional[LLMBackend] = None
    paged_attention_num_blocks: Optional[int] = None

    # scheduler_config
    npu_queue_limit: Optional[int] = None
    max_processing_samples: Optional[int] = None
    spare_blocks_ratio: Optional[float] = None
    is_offline: Optional[bool] = None

    speculative_model_paged_attention_num_blocks: Optional[int] = None
    packing_type: Literal["IDENTITY"] = "IDENTITY"

    @staticmethod
    def add_cli_args(parser: ArgumentParser) -> ArgumentParser:
        """Shared CLI arguments for vLLM engine."""

        # Model arguments
        parser.add_argument(
            '--model',
            type=str,
            required=True,
            help='Path to the LLM engine artifact (Pretrained id will be supported in the future releases).',
        )
        parser.add_argument(
            '--num-speculative-tokens',
            type=int,
            default=EngineArgs.num_speculative_tokens,
            help='The number of tokens that specualtive model will generate speculatively during each iteration of the decoding process.',
        )
        parser.add_argument(
            '--tokenizer',
            type=str,
            default=EngineArgs.tokenizer,
            help='The name or path of a HuggingFace Transformers tokenizer.',
        )
        parser.add_argument(
            '--tokenizer-mode',
            type=str,
            default=EngineArgs.tokenizer_mode,
            help='The tokenizer mode. "auto" will use the fast tokenizer '
            'if available, and "slow" will always use the slow tokenizer.',
        )
        parser.add_argument(
            '--seed',
            type=int,
            default=EngineArgs.seed,
            help='The seed to initialize the random number generator for sampling.',
        )

        # Furiosa LLM specific arguments
        parser.add_argument(
            '--devices',
            type=str,
            default=EngineArgs.devices,
            help='The devices to run the model. It can be a single device or a list of devices. '
            'Each device can be either "npu:X" or "npu:X:*" where X is a specific device index. '
            'If not given, available devices will be used.',
        )
        parser.add_argument(
            '--data-parallel-size',
            type=int,
            default=EngineArgs.data_parallel_size,
            help='The size of the data parallelism group. '
            'If not given, it will be inferred from total avaialble PEs and other parallelism degrees.',
        )
        parser.add_argument(
            '--cache-dir',
            type=Path,
            default=EngineArgs.cache_dir,
            help='The cache directory for temporarily generated files for this LLM instance. '
            'When its value is ``None``, caching is disabled. The default is "$HOME/.cache/furiosa/llm".',
        )
        parser.add_argument(
            '--backend',
            type=str,
            default=EngineArgs.backend,
            help='The backend implementation to run forward() of a model for the LLM. '
            'If not specified, the backend will be chosen based on the device kind.',
        )
        parser.add_argument(
            '--paged-attention-num-blocks',
            type=int,
            default=EngineArgs.paged_attention_num_blocks,
            help='The maximum number of blocks that each k/v storage per layer can store. '
            'This argument must be given if model uses paged attention.',
        )
        parser.add_argument(
            '--npu-queue-limit',
            type=int,
            default=EngineArgs.npu_queue_limit,
            help='The NPU queue limit of the scheduler config.',
        )
        parser.add_argument(
            '--max-processing-samples',
            type=int,
            default=EngineArgs.max_processing_samples,
            help='The maximum processing samples. Used as an hint for the scheduler.',
        )
        parser.add_argument(
            '--spare-blocks-ratio',
            type=float,
            default=EngineArgs.spare_blocks_ratio,
            help='The spare blocks ratio. Used as an hint for the scheduler.',
        )
        parser.add_argument(
            '--is-offline',
            type=bool,
            default=EngineArgs.is_offline,
            help='If True, the scheduler will assume the workload will be offline scenario.',
        )
        parser.add_argument(
            '--use-speculative-decoding-if-possible',
            type=bool,
            default=EngineArgs.use_speculative_decoding_if_possible,
            help="If True, speculative decoding will be used if possible "
            "(`speculative_model` is given or there's artifacts for specualtive model in the artifacts.). "
            "Otherwise, speculative decoding will not be used.",
        )
        parser.add_argument(
            '--speculative-model-paged-attention-num-blocks',
            type=int,
            default=EngineArgs.speculative_model_paged_attention_num_blocks,
            help='The maximum number of blocks that each k/v storage per layer can store for the specualtive model. '
            'This argument must be given if the specualtive model uses paged attention.',
        )
        parser.add_argument(
            '--packing-type',
            type=str,
            default=EngineArgs.packing_type,
            help='Packing algorithm. Possible values are "IDENTITY" only for now',
        )

        return parser

    @classmethod
    def from_cli_args(cls, args: argparse.Namespace):
        attrs = [attr.name for attr in fields(cls)]
        engine_args = cls(**{attr: getattr(args, attr) for attr in attrs})
        return engine_args


@dataclass
class AsyncEngineArgs(EngineArgs):
    # TODO: add async-specific arguments

    @staticmethod
    def add_cli_args(parser: ArgumentParser) -> ArgumentParser:
        # TODO: add async-specific arguments
        parser = EngineArgs.add_cli_args(parser)
        return parser


# XXX: Since SamplingParams.max_tokens in Rust is not an Option type,
# we must ensure max_tokens is not None when SamplingParams is converted from Python to Rust.
# That's why the validation logic is duplicated here and in `LLM._verify_token_len_and_finalize_max_tokens`.
# Unfortunately there is no way to avoid this duplication while minimizing unnecessary encode/decode operations
# and keeping the Python API compatible with vLLM at the same time.
# The best solution would be to change SamplingParams.max_tokens in Rust to an Option type in the future.
# related PR: https://github.com/furiosa-ai/furiosa-runtime/pull/1260
class LLMEngineBase:
    prompt_max_seq_len: int
    max_seq_len_to_capture: int

    def _verify_token_len_and_finalize_max_tokens(
        self, sampling_params: SamplingParams, prompt_len: int
    ):
        if prompt_len > self.prompt_max_seq_len:
            raise ValueError(
                f"This model's maximum input context length is {self.prompt_max_seq_len} tokens, however you requested {prompt_len} tokens. Please reduce your prompt."
            )
        if sampling_params.max_tokens is None:
            sampling_params.max_tokens = self.max_seq_len_to_capture - prompt_len
        else:
            if sampling_params.max_tokens + prompt_len > self.max_seq_len_to_capture:
                # Same error message as OpenAI's response
                raise ValueError(
                    f"This model's maximum context length is {self.max_seq_len_to_capture} tokens, however you requested {sampling_params.max_tokens + prompt_len} tokens ({prompt_len} in your prompt; {sampling_params.max_tokens} for the completion). Please reduce your prompt; or completion length."
                )


class LLMEngine(LLMEngineBase):
    """
    LLMEngine receives requests and generates texts.
    Implements the API interface compatible with vLLM's `LLMEngine`, but this class is based on furiosa-runtime and FuriosaAI NPU.

    The request scheduling approach of this engine is different from that of vLLM's . While vLLM provides
    fine-grained control over decoding via the `step` method, this engine immediately begins
    text generation in the background as soon as a request is submitted via :meth:`add_request`,
    continuing asynchronously until completion. The generated results are placed in a queue that
    clients can retrieve by calling :meth:`step`.

    The Furiosa native engine handles scheduling and batching internally,
    allowing clients to retrieve results via :meth:`step` calls without needing to manage the decoding schedule.

    Please note that cancelling a request using `abort_request` is not supported for now.
    """

    def __init__(
        self,
        native_engine: NativeLLMEngine,
        tokenizer: AnyTokenizer,
        prompt_max_seq_len: int,
        max_seq_len_to_capture: int,
        output_kind: RequestOutputKind = RequestOutputKind.CUMULATIVE,
    ):
        self.native_engine = native_engine
        self.tokenizer = tokenizer
        self.prompt_max_seq_len = prompt_max_seq_len
        self.max_seq_len_to_capture = max_seq_len_to_capture

        # Queue[Tuple[request_id, NativeRequestOutput]]
        self.queue: asyncio.Queue[Tuple[str, NativeRequestOutput]] = asyncio.Queue()
        self.unfinished = 0

        self.output_kind = output_kind
        self.converters: Dict[str, NativeOutputConverter] = {}

        self.aio_loop = asyncio.new_event_loop()

    @classmethod
    def from_llm(
        cls, llm: LLM, output_kind: RequestOutputKind = RequestOutputKind.CUMULATIVE
    ) -> "LLMEngine":
        assert (
            llm.max_seq_len_to_capture is not None
        ), "Generative models must have max_seq_len_to_capture set."
        return cls(
            llm.engine,
            llm.tokenizer,
            llm.prompt_max_seq_len,
            llm.max_seq_len_to_capture,
            output_kind,
        )

    # output_kind is not exposed in EngineArgs because it is used internally
    @classmethod
    def from_engine_args(cls, args: EngineArgs) -> "LLMEngine":
        """
        Creates an LLMEngine from EngineArgs.
        """
        try:
            scheduler_config_ = SchedulerConfig.load(f"{args.model}/scheduler_config.json")
            for scheduler_config_attr in fields(SchedulerConfig):
                if v := getattr(args, scheduler_config_attr.name, None) is not None:
                    setattr(scheduler_config_, scheduler_config_attr.name, v)
        except Exception:
            scheduler_config_ = None

        llm = LLM.from_artifacts(
            path=args.model,
            num_speculative_tokens=args.num_speculative_tokens,
            use_speculative_decoding_if_possible=args.use_speculative_decoding_if_possible,
            data_parallel_size=args.data_parallel_size,
            tokenizer=args.tokenizer,
            tokenizer_mode=args.tokenizer_mode,
            seed=args.seed,
            devices=args.devices,
            cache_dir=args.cache_dir,
            backend=args.backend,
            paged_attention_num_blocks=args.paged_attention_num_blocks,
            scheduler_config=scheduler_config_,
            speculative_model_paged_attention_num_blocks=args.speculative_model_paged_attention_num_blocks,
            packing_type=args.packing_type,
        )

        return cls.from_llm(llm)

    def add_request(
        self,
        request_id: str,
        prompt: PromptType,
        sampling_params: SamplingParams,
    ) -> None:
        """
        Adds a new request to the engine.
        The decoding iteration starts immediately after adding the request.

        Args:
            request_id: The unique id of the request.
            prompt: The prompt to the LLM.
            sampling_params: The sampling parameters of the request.
        """
        prompt_str, batch_encoding = preprocess_prompt(prompt, self.tokenizer)
        prompt_token_ids = batch_encoding["input_ids"]
        self._verify_token_len_and_finalize_max_tokens(sampling_params, len(prompt_token_ids))

        n = sampling_params.n if sampling_params is not None else 1
        self.unfinished += 1
        self.converters[request_id] = NativeOutputConverter(
            self.tokenizer, n, self.output_kind, request_id, prompt_str, prompt_token_ids
        )
        self.aio_loop.create_task(
            self._process_request(request_id, batch_encoding, sampling_params)
        )

    async def _process_request(
        self,
        request_id: str,
        batch_encoding: BatchEncoding,
        sampling_params: Optional[SamplingParams] = None,
    ) -> None:
        async for output in self.native_engine.stream_generate(batch_encoding, sampling_params):
            await self.queue.put((request_id, output))

    def has_unfinished_requests(self) -> bool:
        """
        Returns True if there are unfinished requests.
        """
        return self.unfinished > 0

    def step(self) -> List[RequestOutput]:
        """
        Returns newly generated results of one decoding iteration from the queue.
        """
        output: NativeRequestOutput
        req_id, output = self.aio_loop.run_until_complete(self.queue.get())
        # currently only single output is supported inside `hf_compat_stream_generate`
        # if the behavior is changed, this part should be updated accordingly
        assert len(output.outputs) == 1
        converter = self.converters[req_id]
        req_output = converter.convert(output)

        if req_output is None:
            return []

        if req_output.finished:
            self.unfinished -= 1
            del self.converters[req_id]

        return [req_output]


class AsyncLLMEngine(LLMEngineBase):
    """
    AsyncLLMEngine receives requests and generates texts asynchronously.
    Implements the API interface compatible with vLLM's `AsyncLLMEngine`, but this class is based on furiosa-runtime and FuriosaAI NPU.
    """

    def __init__(
        self,
        native_engine: NativeLLMEngine,
        tokenizer: AnyTokenizer,
        prompt_max_seq_len: int,
        max_seq_len_to_capture: int,
        output_kind: RequestOutputKind = RequestOutputKind.CUMULATIVE,
    ):
        self.native_engine = native_engine
        self.tokenizer = tokenizer
        self.prompt_max_seq_len = prompt_max_seq_len
        self.max_seq_len_to_capture = max_seq_len_to_capture

        self.output_kind = output_kind

    @classmethod
    def from_llm(
        cls, llm: LLM, output_kind: RequestOutputKind = RequestOutputKind.CUMULATIVE
    ) -> "AsyncLLMEngine":
        assert (
            llm.max_seq_len_to_capture is not None
        ), "Generative models must have max_seq_len_to_capture set."
        return cls(
            llm.engine,
            llm.tokenizer,
            llm.prompt_max_seq_len,
            llm.max_seq_len_to_capture,
            output_kind,
        )

    @classmethod
    def from_engine_args(cls, args: AsyncEngineArgs) -> "AsyncLLMEngine":
        """
        Creates an AsyncLLMEngine from AsyncEngineArgs.
        """
        try:
            scheduler_config_ = SchedulerConfig.load(f"{args.model}/scheduler_config.json")
            for scheduler_config_attr in fields(SchedulerConfig):
                if v := getattr(args, scheduler_config_attr.name, None) is not None:
                    setattr(scheduler_config_, scheduler_config_attr.name, v)
        except Exception:
            scheduler_config_ = None

        llm = LLM.from_artifacts(
            path=args.model,
            num_speculative_tokens=args.num_speculative_tokens,
            use_speculative_decoding_if_possible=args.use_speculative_decoding_if_possible,
            data_parallel_size=args.data_parallel_size,
            tokenizer=args.tokenizer,
            tokenizer_mode=args.tokenizer_mode,
            seed=args.seed,
            devices=args.devices,
            cache_dir=args.cache_dir,
            backend=args.backend,
            paged_attention_num_blocks=args.paged_attention_num_blocks,
            scheduler_config=scheduler_config_,
            speculative_model_paged_attention_num_blocks=args.speculative_model_paged_attention_num_blocks,
            packing_type=args.packing_type,
        )

        return cls.from_llm(llm)

    async def generate(
        self,
        prompt: PromptType,
        sampling_params: SamplingParams,
        request_id: str,
    ) -> AsyncGenerator[RequestOutput, None]:
        """
        Generates text completions for a given prompt.

        Args:
            prompt: The prompt to the LLM. See :class:`~PromptType`
                for more details about the format of each input.
            sampling_params: The sampling parameters of the request.
            request_id: The unique id of the request.
        """
        async for output in self.add_request(
            request_id,
            prompt,
            sampling_params,
        ):
            yield output

    async def add_request(
        self,
        request_id: str,
        prompt: PromptType,
        sampling_params: SamplingParams,
    ) -> AsyncGenerator[RequestOutput, None]:
        # XXX(n0gu): not exposed in the document level, as vLLM does not
        prompt_str, batch_encoding = preprocess_prompt(prompt, self.tokenizer)
        prompt_token_ids = batch_encoding["input_ids"]
        self._verify_token_len_and_finalize_max_tokens(sampling_params, len(prompt_token_ids))

        converter = NativeOutputConverter(
            self.tokenizer,
            sampling_params.n,
            self.output_kind,
            request_id,
            prompt_str,
            prompt_token_ids,
        )
        async for native_output in self.native_engine.stream_generate(
            batch_encoding, sampling_params
        ):
            output = converter.convert(native_output)
            if output is not None:
                yield output

    # TODO
    # async def engine_step(self): ...
    # async def abort(self, request_id): ...


def preprocess_prompt(
    prompt: PromptType, tokenizer: PreTrainedTokenizerBase
) -> Tuple[str, BatchEncoding]:
    if isinstance(prompt, str):
        prompt_str = prompt
    elif isinstance(prompt, dict):
        if "prompt" in prompt:
            prompt_str = prompt["prompt"]  # type: ignore
        elif "prompt_token_ids" in prompt:
            # To get BatchEncoding, decode first and then encode again.
            # This is inefficient but since generator takes BatchEncoding as input this is the safest way.
            prompt_str = tokenizer.decode(prompt["prompt_token_ids"], skip_special_tokens=True)
        else:
            raise ValueError(f"Unsupported prompt type: {type(prompt)}")
    else:
        raise ValueError(f"Unsupported prompt type: {type(prompt)}")
    return prompt_str, tokenizer(prompt_str, padding=False, add_special_tokens=False)
