from enum import Enum
import logging
from typing import List, Optional, Tuple

from transformers import PreTrainedTokenizerBase


class CompletionOutput:
    """The output data of one completion output of a request.

    Args:
        index: The index of the output in the request.
        text: The generated output text.
        token_ids: The token IDs of the generated output text.
        finish_reason: The reason why the sequence is finished.
    """

    def __init__(
        self,
        index: int,
        text: str,
        token_ids: List[int],
        finish_reason: Optional[str] = None,
    ) -> None:
        self.index = index
        self.text = text
        self.token_ids = token_ids
        self.finish_reason = finish_reason

    def __repr__(self) -> str:
        return (
            f"CompletionOutput(index={self.index}, "
            f"text={self.text!r}, "
            f"token_ids={self.token_ids}, "
            f"finish_reason={self.finish_reason})"
        )


class RequestOutput:
    def __init__(
        self,
        request_id: str,
        prompt: str,
        prompt_token_ids: List[int],
        outputs: List[CompletionOutput],
        finished: bool,
    ) -> None:
        self.request_id = request_id
        self.prompt = prompt
        self.prompt_token_ids = prompt_token_ids
        self.outputs = outputs
        self.finished = finished

    def __repr__(self) -> str:
        return (
            f"RequestOutput(request_id={self.request_id}, "
            f"prompt={self.prompt!r}, "
            f"prompt_token_ids={self.prompt_token_ids}, "
            f"outputs={self.outputs}, "
            f"finished={self.finished})"
        )


class RequestOutputKind(Enum):
    # Return entire output so far in every RequestOutput
    CUMULATIVE = 0
    # Return only deltas in each RequestOutput
    DELTA = 1


class NativeOutputConverter:
    """
    This class has two main functions:
    1. Convert the NativeRequestOutput to RequestOutput, respecting the RequestOutputKind.
    2. For streaming RequestOutput, detokenize incrementally so that
       incomplete tokens (which are represented as replacement character "�") is not returned to the user.
    This class must be initialized per prompt, not per request.

    Args:
        tokenizer: The tokenizer used to tokenize the prompt.
        n: The number of expected outputs. Usually equals to `SamplingParams.n` if not beam search.
        output_kind: Determines how to handle the output. See `RequestOutputKind`.
        request_id: The request ID.
        prompt: The prompt text.
        prompt_token_ids: The prompt token IDs.
    """

    def __init__(
        self,
        tokenizer: PreTrainedTokenizerBase,
        n: int,
        output_kind: RequestOutputKind,
        request_id: str,
        prompt: str,
        prompt_token_ids: List[int],
    ):
        self.tokenizer = tokenizer
        self.output_kind = output_kind
        self.request_id = request_id
        self.prompt = prompt
        self.prompt_token_ids = prompt_token_ids
        self.detokenizers = [IncrementalDetokenizer(tokenizer, output_kind, True) for _ in range(n)]

    def convert(
        self,
        output: "NativeRequestOutput",  # type: ignore # noqa: F821
    ) -> Optional[RequestOutput]:
        completion_outputs: List[CompletionOutput] = []
        for o in output.outputs:
            detokenizer = self.detokenizers[o.index]
            is_final = o.finish_reason is not None
            text, token_ids = detokenizer.detokenize(o.token_ids, is_final)

            if not token_ids and o.finish_reason is None:
                # streaming mode with incomplete character (that ends with "�")
                continue

            completion_outputs.append(CompletionOutput(o.index, text, token_ids, o.finish_reason))

        if not completion_outputs:
            return None

        return RequestOutput(
            request_id=self.request_id,
            prompt=self.prompt,
            prompt_token_ids=self.prompt_token_ids,
            outputs=completion_outputs,
            finished=all(o.finish_reason is not None for o in output.outputs),
        )


class IncrementalDetokenizer:
    # Maximum number of tokenizer.decode() calls allowed per completion output.
    # Set to 4 since BPE-based tokenizers typically split natural language words into at most 4 tokens.
    MAX_DECODE_TRAIL = 4

    def __init__(
        self,
        tokenizer: PreTrainedTokenizerBase,
        output_kind: RequestOutputKind,
        skip_special_tokens: bool = True,
    ):
        self.tokenizer = tokenizer
        self.output_kind = output_kind
        self.skip_special_tokens = skip_special_tokens
        # Token ids that are buffered but not yet returned
        self.buffered_token_ids: List[int] = []

        # Returned text so far (for RequestOutputKind.CUMULATIVE)
        self.text = ""
        # Returned token ids so far (for RequestOutputKind.CUMULATIVE)
        self.token_ids: List[int] = []

    def detokenize(self, new_token_ids: List[int], is_final: bool) -> Tuple[str, List[int]]:
        """
        Decodes the complete sequence of token IDs into text.

        Although this function is intended to be the final processing step,
        we clear self.tokens as a safeguard against subsequent calls
        """
        if is_final:
            self.buffered_token_ids.extend(new_token_ids)
            delta_token_ids = self.buffered_token_ids
            delta_text = self._decode(delta_token_ids)
        else:
            if len(new_token_ids) > self.MAX_DECODE_TRAIL:
                # This scenario is unlikely since for streaming requests (is_final=False)
                # the Generator normally emits a CompletionOutput per individual token.
                # Multiple tokens returned in a single CompletionOutput could negatively impact
                # TPOT and user experience.
                logging.warning(
                    f"Received {len(new_token_ids)} tokens at once in streaming mode. "
                    f"This may degrade TPOT and user experience. "
                    f"Keeping only last {self.MAX_DECODE_TRAIL} tokens."
                )
                self.buffered_token_ids.extend(new_token_ids[: -self.MAX_DECODE_TRAIL])
                new_token_ids = new_token_ids[-self.MAX_DECODE_TRAIL :]
            delta_token_ids = []
            delta_text = ""
            for new_token_id in new_token_ids:
                decoded = self._append_and_try_detokenize(new_token_id)
                if decoded is None:
                    continue
                text, token_ids = decoded
                delta_text += text
                delta_token_ids.extend(token_ids)

        if self.is_delta:
            return delta_text, delta_token_ids
        else:
            self.text += delta_text
            self.token_ids.extend(delta_token_ids)
            return self.text, self.token_ids

    @property
    def is_delta(self) -> bool:
        return self.output_kind == RequestOutputKind.DELTA

    def _decode(self, token_ids: List[int]) -> str:
        return self.tokenizer.decode(
            token_ids,
            skip_special_tokens=self.skip_special_tokens,
            cleanup_tokenization_spaces=True,
        )

    def _append_and_try_detokenize(self, new_token_id: int) -> Optional[Tuple[str, List[int]]]:
        self.buffered_token_ids.append(new_token_id)
        text = self._decode(self.buffered_token_ids)
        if text.endswith("�"):
            return None
        token_ids = self.buffered_token_ids[:]
        self.buffered_token_ids.clear()
        return text, token_ids
